module Form.Select.Internal exposing
    ( State, ViewState
    , init, initialViewState
    , Msg, update
    , render
    , reInitialise, reset
    , setDefaultLabel, setToLabel
    , setInitialOption, setSelectedOption, setIsOptionDisabled
    , setIsError, setIsLocked, setIsClearable
    , setId
    , getIsChanged, getIsOpen
    , getSelectedOption, getInitialOption
    , getId
    )

import Html exposing (..)
import Html.Lazy exposing (..)
import Html.Events exposing (..)
import Html.Attributes as Attributes exposing (..)

import Dom
import Dict
import Task

import List.Extra as List

import Json.Decode as Decode exposing (Decoder)

import Form.Helpers as Form
import Html.Bdt as Html exposing ((?))
import Resettable exposing (Resettable)


-- MODEL --


type alias State option =
    { isOpen : Bool
    , options : List option
    , selectedOption : Resettable (Maybe option)
    , focusedOption : Maybe option
    }


init : List option -> State option
init options =
    { isOpen = False
    , options = options
    , selectedOption = Resettable.init Nothing
    , focusedOption = Nothing
    }


type alias ViewState option =
    { isLocked : Bool
    , isClearable : Bool
    , isError : Bool
    , isOptionDisabled : option -> Bool
    , toLabel : option -> String
    , defaultLabel : String
    , id : Maybe String
    }


initialViewState : ViewState option
initialViewState =
    { isLocked = False
    , isClearable = False
    , isError = False
    , isOptionDisabled = always False
    , toLabel = toString
    , defaultLabel = "-- Nothing Selected --"
    , id = Nothing
    }


-- UPDATE --


type Msg option
    = Open
    | Blur
    | Select option
    | Clear
    | KeyboardInput Bool KeyboardInput
    | Focus option
    | BlurOption option
    | DomFocus (Result Dom.Error ())


update : Msg option -> State option -> (State option, Cmd (Msg option))
update msg state =

    case msg of
        Open ->
            { state | isOpen = True } ! []

        Blur ->
            case state.focusedOption of
                Nothing ->
                    { state | isOpen = False, focusedOption = Nothing } ! []
                _ ->
                    state ! []

        Select option ->
            { state
                | selectedOption = Resettable.update (Just option) state.selectedOption
                , isOpen = False
                , focusedOption = Nothing
            } ! []

        Clear ->
            { state | selectedOption = Resettable.update Nothing state.selectedOption } ! []

        KeyboardInput isOptionDisabled keyboardInput ->
            handleKeyboardInput state isOptionDisabled keyboardInput

        Focus option ->
            { state | focusedOption = Just (focusOption state.options option) } ! []

        BlurOption option ->
            case state.focusedOption == Just option of
                True ->
                    { state | focusedOption = Nothing, isOpen = False } ! []

                False ->
                    state ! []

        DomFocus _ ->
            state ! []


type KeyboardInput
    = Enter
    | Space
    | Up
    | Down


onKeyboardInput : (KeyboardInput -> msg) -> Attribute msg
onKeyboardInput msg =

    onWithOptions
        "keydown"
        { preventDefault = True, stopPropagation = True }
        (Decode.andThen (isSelectInputKey msg) keyCode)


isSelectInputKey : (KeyboardInput -> msg) -> Int -> Decoder msg
isSelectInputKey msg code =

    let
        dict =
            Dict.fromList [(13, Enter), (32, Space), (38, Up), (40, Down)]

    in
        case Dict.get code dict of
            Just selectKeyboardInput ->
                Decode.succeed (msg selectKeyboardInput)

            Nothing ->
                Decode.fail "Not a select input key"


handleKeyboardInput: State option -> Bool -> KeyboardInput -> (State option, Cmd (Msg option))
handleKeyboardInput state isOptionDisabled keyboardInput =

    case state.focusedOption of

        Nothing ->
            { state | focusedOption = List.head state.options } ! []

        Just focusedOption ->
            case keyboardInput of
                Enter ->
                    case isOptionDisabled of
                        True ->
                            state ! []

                        False ->
                            { state
                                | selectedOption = Resettable.update (Just focusedOption) state.selectedOption
                                , isOpen = False
                            } ! []

                Space ->
                    case isOptionDisabled of
                        True ->
                            state ! []

                        False ->
                            { state
                                | selectedOption = Resettable.update (Just focusedOption) state.selectedOption
                                , isOpen = False
                            } ! []

                Up ->
                    let
                        newFocusedOption =
                            focusPreviousOption state.options focusedOption

                    in
                        { state | focusedOption = Just newFocusedOption }
                            ! [ Task.attempt DomFocus (Form.toHtmlId newFocusedOption |> Dom.focus) ]

                Down ->
                    let
                        newFocusedOption =
                            focusNextOption state.options focusedOption

                    in
                        { state | focusedOption = Just newFocusedOption }
                            ! [ Task.attempt DomFocus (Form.toHtmlId newFocusedOption |> Dom.focus) ]


focusOption : List option -> option -> option
focusOption options option =

    case List.member option options of
        True ->
            option

        False ->
            Debug.crash ("SELECT ERROR - can't focus" ++ Form.toHtmlId option ++ " it is not a valid option for this select.")


focusPreviousOption : List option -> option -> option
focusPreviousOption options option =

    focusNextOption (List.reverse options) option


focusNextOption : List option -> option -> option
focusNextOption options option =

    options
        |> List.dropWhile ((/=) option)
        |> List.drop 1
        |> List.head
        |> Maybe.withDefault option


-- VIEW --


render : State option -> ViewState option -> Html (Msg option)
render state viewState =

    case state.isOpen of
        False ->
            lazy2 closed state viewState

        True ->
            lazy2 open state viewState


closed : State option -> ViewState option -> Html (Msg option)
closed state viewState =

    div
        [ class "bdt-elm select-container" ]
        [ div
            [ class "input"
            , classList [("locked", viewState.isLocked), ("error", viewState.isError)]
            , Html.maybeAttribute id viewState.id
            , disabled viewState.isLocked
            , tabindex 0 ? not viewState.isLocked
            , onFocus Open ? not viewState.isLocked
            , onClick Open ? not viewState.isLocked
            ]
            [ div
                [ class "option-label"
                , title (Maybe.map viewState.toLabel (Resettable.getValue state.selectedOption) |> Maybe.withDefault viewState.defaultLabel)
                ]
                [ text (Maybe.map viewState.toLabel (Resettable.getValue state.selectedOption) |> Maybe.withDefault viewState.defaultLabel) ]
            , clearButton state viewState
            , Html.divIf (not viewState.isLocked)
                [ class "material-icons" ]
                [ text "expand_more" ]
            ]
        ]


open : State option -> ViewState option -> Html (Msg option)
open state viewState =

    div
        [ class "bdt-elm select-container" ]
        [ div
            [ class "input"
            , classList [("locked", viewState.isLocked), ("error", viewState.isError)]
            , Html.maybeAttribute id viewState.id
            , onKeyboardInput <| KeyboardInput False
            , tabindex -1
            , onBlur Blur
            ]
            [ div
                [ class "option-label"
                , title (Maybe.map viewState.toLabel (Resettable.getValue state.selectedOption) |> Maybe.withDefault viewState.defaultLabel)
                ]
                [ text (Maybe.map viewState.toLabel (Resettable.getValue state.selectedOption) |> Maybe.withDefault viewState.defaultLabel) ]
            ]
        , optionList state viewState
        ]


clearButton : State option -> ViewState option -> Html (Msg option)
clearButton state viewState =

    Html.divIf (viewState.isClearable && Resettable.getValue state.selectedOption /= Nothing)
        [ class "material-icons"
        , onWithOptions "mousedown" { preventDefault = True, stopPropagation = True } (Decode.succeed Clear)
        ]
        [ text "clear" ]


optionList : State option -> ViewState option -> Html (Msg option)
optionList state viewState =

    div
        [ class "option-list" ]
        (List.map (optionItem state viewState) state.options)


optionItem : State option -> ViewState option -> option -> Html (Msg option)
optionItem state viewState option =

    div
        [ class "option-item"
        , classList [("disabled", viewState.isOptionDisabled option), ("focus", state.focusedOption == Just option)]
        , id <| Form.toHtmlId option
        , onFocus <| Focus option
        , onBlur <| BlurOption option
        , onKeyboardInput <| KeyboardInput (viewState.isOptionDisabled option)
        , tabindex -1
        , onMouseDown (Select option) ? not (viewState.isOptionDisabled option)
        ]
        [ text (viewState.toLabel option) ]


-- STATE SETTERS --


reInitialise : State option -> State option
reInitialise state =

    { state | selectedOption = Resettable.init (Resettable.getValue state.selectedOption) }


reset : State option -> State option
reset state =

    { state | selectedOption = Resettable.reset state.selectedOption }


setInitialOption : Maybe option -> State option -> State option
setInitialOption selectedOption state =

    { state | selectedOption = Resettable.init selectedOption }


setSelectedOption : Maybe option -> State option -> State option
setSelectedOption selectedOption state =

    { state | selectedOption = Resettable.update selectedOption state.selectedOption }


-- VIEW STATE SETTERS --


setToLabel : (option -> String) -> ViewState option -> ViewState option
setToLabel toLabel viewState =

    { viewState | toLabel = toLabel }


setDefaultLabel : String -> ViewState option -> ViewState option
setDefaultLabel defaultLabel viewState =

    { viewState | defaultLabel = defaultLabel }


setIsOptionDisabled : (option -> Bool) -> ViewState option -> ViewState option
setIsOptionDisabled isOptionDisabled viewState =

    { viewState | isOptionDisabled = isOptionDisabled }


setIsLocked : Bool -> ViewState option -> ViewState option
setIsLocked isLocked viewState =

    { viewState | isLocked = isLocked }


setIsError : Bool -> ViewState option -> ViewState option
setIsError isError viewState =

    { viewState | isError = isError }


setIsClearable : Bool -> ViewState option -> ViewState option
setIsClearable isClearable viewState =

    { viewState | isClearable = isClearable }


setId : String -> ViewState option -> ViewState option
setId id viewState =

    { viewState | id = Just id }


-- GETTERS --


getIsChanged : State option -> Bool
getIsChanged state =

    Resettable.getIsChanged state.selectedOption


getInitialOption : State option -> Maybe option
getInitialOption state =

    Resettable.getInitialValue state.selectedOption


getSelectedOption : State option -> Maybe option
getSelectedOption state =

    Resettable.getValue state.selectedOption


getIsOpen : State option -> Bool
getIsOpen =
    .isOpen


getId : ViewState option -> Maybe String
getId =
    .id